import operator

import pytest

from pandas.core import ops

from pandas import Series
import numpy as np

from ladybug.epw import EPW

from ladybug_pandas.extension_types.dtype import LadybugDType
from ladybug_pandas.extension_types.arraytype import LadybugArrayType


@pytest.fixture(scope='session')
def dbt_data_collection():
    epw_path = 'tests/assets/epw/tokyo.epw'

    epw = EPW(epw_path)

    return epw.dry_bulb_temperature


@pytest.fixture(scope='session')
def dtype(dbt_data_collection):
    """A fixture providing the ExtensionDtype to validate."""
    return LadybugDType.construct_from_header(dbt_data_collection.header)


# @pytest.fixture
# def dbt_array(dtype):
#     return LadybugArrayType

@pytest.fixture
def data(dbt_data_collection, dtype):
    """
    Length-100 array for this type.
    * data[0] and data[1] should both be non missing
    * data[0] and data[1] should not be equal
    """

    # values = dbt_data_collection.values[:100]

    # return dbt_array(values)
    # dbt_data_collection.values = dbt_data_collection[:100]

    return LadybugArrayType(dbt_data_collection.values[:100], dtype=dtype)


@pytest.fixture
def data_for_twos():
    """Length-100 array in which all the elements are two."""
    raise NotImplementedError


@pytest.fixture
def data_missing(dtype):
    """Length-2 array with [NA, Valid]"""
    return LadybugArrayType([np.nan, 3.4], dtype=dtype)


@pytest.fixture(params=["data", "data_missing"])
def all_data(request, data, data_missing):
    """Parametrized fixture giving 'data' and 'data_missing'"""
    if request.param == "data":
        return data
    elif request.param == "data_missing":
        return data_missing


@pytest.fixture
def data_repeated(data):
    """
    Generate many datasets.
    Parameters
    ----------
    data : fixture implementing `data`
    Returns
    -------
    Callable[[int], Generator]:
        A callable that takes a `count` argument and
        returns a generator yielding `count` datasets.
    """

    def gen(count):
        for _ in range(count):
            yield data

    return gen


@pytest.fixture
def data_for_sorting(dtype):
    """
    Length-3 array with a known sort order.
    This should be three items [B, C, A] with
    A < B < C
    """
    return LadybugArrayType([5, 10, 2], dtype=dtype)


@pytest.fixture
def data_missing_for_sorting(dtype):
    """
    Length-3 array with a known sort order.
    This should be three items [B, NA, A] with
    A < B and NA missing.
    """
    return LadybugArrayType([5, None, 2], dtype=dtype)


@pytest.fixture
def data_for_twos(dtype):
    """Length-100 array in which all the elements are two."""
    return LadybugArrayType([2]*100, dtype=dtype)

@pytest.fixture
def na_cmp():
    """
    Binary operator for comparing NA values.
    Should return a function of two arguments that returns
    True if both arguments are (scalar) NA for your type.
    By default, uses ``operator.is_``
    """
    def comp(a, b):
        return np.isnan(a) and np.isnan(b)

    return comp
    # return operator.is_


@pytest.fixture
def na_value():
    """The scalar missing value for this type. Default 'None'"""
    return np.nan


@pytest.fixture
def data_for_grouping(dtype):
    """
    Data for factorization, grouping, and unique tests.
    Expected to be like [B, B, NA, NA, A, A, B, C]
    Where A < B < C and NA is missing
    """
    return LadybugArrayType([3, 3, None, None, 1, 1, 3, 5], dtype=dtype)


@pytest.fixture(params=[True, False])
def box_in_series(request):
    """Whether to box the data in a Series"""
    return request.param


@pytest.fixture(
    params=[
        lambda x: 1,
        lambda x: [1] * len(x),
        lambda x: Series([1] * len(x)),
        lambda x: x,
    ],
    ids=["scalar", "list", "series", "object"],
)
def groupby_apply_op(request):
    """
    Functions to test groupby.apply().
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_frame(request):
    """
    Boolean fixture to support Series and Series.to_frame() comparison testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_series(request):
    """
    Boolean fixture to support arr and Series(arr) comparison testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def use_numpy(request):
    """
    Boolean fixture to support comparison testing of ExtensionDtype array
    and numpy array.
    """
    return request.param


@pytest.fixture(params=["ffill", "bfill"])
def fillna_method(request):
    """
    Parametrized fixture giving method parameters 'ffill' and 'bfill' for
    Series.fillna(method=<method>) testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_array(request):
    """
    Boolean fixture to support ExtensionDtype _from_sequence method testing.
    """
    return request.param


# ----------------------------------------------------------------
# Operators & Operations
# ----------------------------------------------------------------
_all_arithmetic_operators = [
    "__add__",
    # "__radd__",
    "__sub__",
    # "__rsub__",
    "__mul__",
    # "__rmul__",
    "__floordiv__",
    # "__rfloordiv__",
    "__truediv__",
    # "__rtruediv__",
    "__pow__",
    # "__rpow__",
    "__mod__",
    # "__rmod__",
]


@pytest.fixture(params=_all_arithmetic_operators)
def all_arithmetic_operators(request):
    """
    Fixture for dunder names for common arithmetic operations.
    """
    return request.param


@pytest.fixture(
    params=[
        operator.add,
        ops.radd,
        operator.sub,
        ops.rsub,
        operator.mul,
        ops.rmul,
        operator.truediv,
        ops.rtruediv,
        operator.floordiv,
        ops.rfloordiv,
        operator.mod,
        ops.rmod,
        operator.pow,
        ops.rpow,
    ]
)
def all_arithmetic_functions(request):
    """
    Fixture for operator and roperator arithmetic functions.

    Notes
    -----
    This includes divmod and rdivmod, whereas all_arithmetic_operators
    does not.
    """
    return request.param


_all_numeric_reductions = [
    "sum",
    "max",
    "min",
    "mean",
    "prod",
    "std",
    "var",
    "median",
    "kurt",
    "skew",
]


@pytest.fixture(params=_all_numeric_reductions)
def all_numeric_reductions(request):
    """
    Fixture for numeric reduction names.
    """
    return request.param


_all_boolean_reductions = ["all", "any"]


@pytest.fixture(params=_all_boolean_reductions)
def all_boolean_reductions(request):
    """
    Fixture for boolean reduction names.
    """
    return request.param


@pytest.fixture(params=["__eq__", "__ne__", "__le__", "__lt__", "__ge__", "__gt__"])
def all_compare_operators(request):
    """
    Fixture for dunder names for common compare operations

    * >=
    * >
    * ==
    * !=
    * <
    * <=
    """
    return request.param


@pytest.fixture(params=["__le__", "__lt__", "__ge__", "__gt__"])
def compare_operators_no_eq_ne(request):
    """
    Fixture for dunder names for compare operations except == and !=

    * >=
    * >
    * <
    * <=
    """
    return request.param


@pytest.fixture(
    params=["__and__", "__rand__", "__or__", "__ror__", "__xor__", "__rxor__"]
)
def all_logical_operators(request):
    """
    Fixture for dunder names for common logical operations

    * |
    * &
    * ^
    """
    return request.param




import pytest

import pandas as pd

from ladybug.datatype.temperature import DryBulbTemperature
from ladybug.datatype.energyintensity import DirectNormalRadiation
from ladybug.epw import EPW

from ladybug_pandas.extension_types.dtype import LadybugDType
from ladybug_pandas.extension_types.arraytype import LadybugArrayType

from ladybug_pandas import dataframe_from_collections, dataframe_from_epw

@pytest.fixture()
def epw():
    epw_path = 'tests/assets/epw/tokyo.epw'

    epw = EPW(epw_path)

    epw._import_data()

    return epw

@pytest.fixture()
def temp_series(epw):
    return pd.Series(LadybugArrayType._from_data_collection(epw.dry_bulb_temperature))

@pytest.fixture()
def dnr_series(epw):
    return pd.Series(LadybugArrayType._from_data_collection(epw.direct_normal_radiation))


def test_addition(temp_series, dnr_series):
    res = temp_series + dnr_series

    # should not raise


def test_conversion(temp_series):

    array = temp_series.values

    new_array = array.to_ip()

    # should not raise


def test_from_collections(epw):

    df = dataframe_from_collections([epw.dry_bulb_temperature, epw.direct_normal_radiation])

    # should not raise

def test_from_epw(epw):

    df = dataframe_from_epw(epw)

    # should not raise


import re
import types
from typing import Sequence, Any
from pandas.api.extensions import ExtensionArray, ExtensionDtype,register_extension_dtype
from pandas._typing import ArrayLike
import pandas as pd
import numpy as np

from ladybug.datatype.base import DataTypeBase
from ladybug.header import Header
from ladybug.datatype import TYPESDICT, UNITS


def _to_unit_base_hack(self, base_unit, values, unit, from_unit):
    """Return values in a given unit given the input from_unit."""
    namespace = {'self': self, 'values': values}
    if not from_unit == base_unit:
        self.is_unit_acceptable(from_unit, True)
        statement = 'self._{}_to_{}(values)'.format(
            self._clean(from_unit), self._clean(base_unit))
        values = eval(statement, namespace)
    if not unit == base_unit:
        self.is_unit_acceptable(unit, True)
        statement = 'self._{}_to_{}(values)'.format(
            self._clean(base_unit), self._clean(unit))
        values = eval(statement, namespace)
    return values

def _is_numeric_hack(*args, **kwargs):
    return True

funcType = types.MethodType

@register_extension_dtype
class LadybugDType(ExtensionDtype):

    _metadata = ('data_type_name', 'unit')

    def __init__(self, data_type=None, unit=None):

        if data_type is None and unit is None:
            self.data_type = DataTypeBase('Undefined Type')
            self.unit = 'None'
        else:
            assert isinstance(data_type, DataTypeBase), ValueError(f'{data_type} is not an instance of type DataTypeBase')
            assert unit in data_type._units, ValueError(f'unit {unit} should be one of {data_type.units}')

            # Replace to unit base function to work with numpy arrays

            self.data_type = data_type
            self.unit = unit

        # setattr(self.data_type, '_is_numeric', _is_numeric_hack)
        # self.data_type._to_unit_base = _to_unit_base_hack.__get__()
        self.data_type._to_unit_base = funcType(_to_unit_base_hack, self.data_type)

    @classmethod
    def construct_from_string(cls, string):
        if not isinstance(string, str):
            raise TypeError(
                "'construct_from_string' expects a string, got {}".format(type(string))
            )

        match = re.fullmatch(r"(.*) \((.*)\)", string)

        if match is None:
            raise TypeError(
                "Cannot construct a 'LadybugDType' from '{}'".format(string)
            )

        type_name = ''.join(match[1].split(' '))
        unit = match[2]

        assert type_name in TYPESDICT, ValueError(f'Ladybug Type of {type_name} is not recognized')
        data_type = TYPESDICT[type_name]()

        assert unit in data_type._units, ValueError(f'Ladybug Type unit of {unit} is not recognized for type {type_name}')

        return cls(data_type=data_type, unit=unit)

    @classmethod
    def construct_array_type(cls):
        from .arraytype import LadybugArrayType

        return LadybugArrayType

    @classmethod
    def construct_from_header(cls, header: Header):
        """[summary]

        Arguments:
            header {Header} -- [description]

        Returns:
            [type] -- [description]
        """
        return cls(data_type=header.data_type, unit=header.unit)

    def to_header(self):
        """Cast the ExtensionDtyp to a ladybug header object

        Returns:
            Header -- A ladybug header object
        """
        return Header(
            data_type=self.data_type,
            unit=self.unit
        )

    @property
    def type(self):
        return np.float

    @property
    def _is_numeric(self):
        return True

    @property
    def kind(self):
        return 'f'

    @property
    def name(self):
        # return "{} ({})".format(self.data_type, self.unit)
        return self.__repr__()

    @property
    def data_type_name(self):
        return self.data_type.name

    def __repr__(self):
        return "{} ({})".format(self.data_type_name, self.unit)


import pytest
from pandas.tests.extension import base
import pandas as pd
from pandas.core import ops


class TestCasting(base.BaseCastingTests):
    pass

class TestDtypes(base.BaseDtypeTests):
    
    def test_array_type(self, data, dtype):
        """Replace with isinstance test because of inheritance"""
        assert isinstance(data, dtype.construct_array_type())
        # assert dtype.construct_array_type() is type(data)

class TestConstructors(base.BaseConstructorsTests):
    pass

class TestGetItem(base.BaseGetitemTests):
    pass

class TestBaseGroupby(base.BaseGroupbyTests):
    pass

class TestBaseInterface(base.BaseInterfaceTests):
    pass

class TestBaseParsing(base.BaseParsingTests):
    pass

class TestBaseMethods(base.BaseMethodsTests):
    pass

class TestBaseMissing(base.BaseMissingTests):
    pass

class TestBaseArithmeticOps(base.BaseArithmeticOpsTests):

    series_scalar_exc = None
    frame_scalar_exc = None
    series_array_exc = None
    divmod_exc = None

    def test_error(self):
        pass

    def test_divmod_series_array(self, data, data_for_twos):
        s = pd.Series(data)
        self._check_divmod_op(s, divmod, data, exc=self.divmod_exc)

        other = data_for_twos
        self._check_divmod_op(other, ops.rdivmod, s, exc=self.divmod_exc)

        other = pd.Series(other)
        self._check_divmod_op(other, ops.rdivmod, s, exc=self.divmod_exc)

# class TestBaseComparisonOps(base.BaseComparisonOpsTests):

#     def _compare_other(self, s, data, op_name, other):
#         op = self.get_op_from_name(op_name)
#         if op_name == "__eq__":
#             assert getattr(data, op_name)(other) is NotImplemented
#             assert not op(s, other).all()
#         elif op_name == "__ne__":
#             assert getattr(data, op_name)(other) is NotImplemented
#             assert op(s, other).all()

#         else:

#             # array
#             assert getattr(data, op_name)(other) is NotImplemented

#             # series
#             s = pd.Series(data)
#             with pytest.raises(TypeError):
#                 op(s, other)


class TestBasePrinting(base.BasePrintingTests):
    pass

class TestBaseBooleanReduce(base.BaseBooleanReduceTests):
    pass

class TestBaseNumericReduce(base.BaseNumericReduceTests):
    pass

class TestBaseReshaping(base.BaseReshapingTests):

    def test_concat_mixed_dtypes(self, data):
        # https://github.com/pandas-dev/pandas/issues/20762
        df1 = pd.DataFrame({"A": data[:3]})
        df2 = pd.DataFrame({"A": [1, 2, 3]})
        df3 = pd.DataFrame({"A": ["a", "b", "c"]}).astype("category")
        dfs = [df1, df2, df3]
    
        # dataframes
        result = pd.concat(dfs)
        expected = pd.concat([x.astype(object) for x in dfs])
        self.assert_frame_equal(result, expected)
    
        # series
        result = pd.concat([x["A"] for x in dfs])
        expected = pd.concat([x["A"].astype(object) for x in dfs])
        self.assert_series_equal(result, expected)
    
        # simple test for just EA and one other
        result = pd.concat([df1, df2])
        expected = pd.concat([df1.astype("object"), df2.astype("object")])
        self.assert_frame_equal(result, expected)
    
        # Commented out the section below because it was failing
        # I think it fails because the ndarray type of the ladybug tools
        # array type is a float which concats into a float rather than an object
        # when presented with a list of integers

        # result = pd.concat([df1["A"], df2["A"]])
        # expected = pd.concat([df1["A"].astype("object"), df2["A"].astype("object")])
        # self.assert_series_equal(result, expected)

    pass

class TestBaseSetitem(base.BaseSetitemTests):
    pass




import operator

from pandas.api.extensions import ExtensionArray, ExtensionDtype, ExtensionScalarOpsMixin
from pandas.core.arrays.base import try_cast_to_ea
from pandas.core import ops
from pandas.compat import set_function_name
from pandas.core.dtypes.generic import ABCExtensionArray, ABCIndexClass, ABCSeries
from pandas.core.dtypes.common import is_array_like, is_list_like

from typing import Sequence, Any, Union, List
from pandas.core.dtypes.generic import ABCExtensionArray

from pandas._typing import ArrayLike
import numpy as np
import pandas as pd

from scipy import stats

from pandas.core.arrays.numpy_ import PandasArray

from ladybug.datacollection import BaseCollection, HourlyDiscontinuousCollection
from ladybug.dt import DateTime

from .dtype import LadybugDType


COMPARISON_OPS = ['eq', 'ne', 'lt', 'gt', 'le', 'ge']

class LadybugExtensionScalarOpsMixin(ExtensionScalarOpsMixin):
    """
    A mixin for defining  ops on an ExtensionArray.

    It is assumed that the underlying scalar objects have the operators
    already defined.

    Notes
    -----
    If you have defined a subclass MyExtensionArray(ExtensionArray), then
    use MyExtensionArray(ExtensionArray, ExtensionScalarOpsMixin) to
    get the arithmetic operators.  After the definition of MyExtensionArray,
    insert the lines

    MyExtensionArray._add_arithmetic_ops()
    MyExtensionArray._add_comparison_ops()

    to link the operators to your class.

    .. note::

       You may want to set ``__array_priority__`` if you want your
       implementation to be called when involved in binary operations
       with NumPy arrays.
    """

    @classmethod
    def _create_method(cls, op, coerce_to_dtype=True):
        """
        A class method that returns a method that will correspond to an
        operator for an ExtensionArray subclass, by dispatching to the
        relevant operator defined on the individual elements of the
        ExtensionArray.

        Parameters
        ----------
        op : function
            An operator that takes arguments op(a, b)
        coerce_to_dtype : bool, default True
            boolean indicating whether to attempt to convert
            the result to the underlying ExtensionArray dtype.
            If it's not possible to create a new ExtensionArray with the
            values, an ndarray is returned instead.

        Returns
        -------
        Callable[[Any, Any], Union[ndarray, ExtensionArray]]
            A method that can be bound to a class. When used, the method
            receives the two arguments, one of which is the instance of
            this class, and should return an ExtensionArray or an ndarray.

            Returning an ndarray may be necessary when the result of the
            `op` cannot be stored in the ExtensionArray. The dtype of the
            ndarray uses NumPy's normal inference rules.

        Examples
        --------
        Given an ExtensionArray subclass called MyExtensionArray, use

        >>> __add__ = cls._create_method(operator.add)

        in the class definition of MyExtensionArray to create the operator
        for addition, that will be based on the operator implementation
        of the underlying elements of the ExtensionArray
        """

        def _binop(self, other):
            def convert_values(param):
                if isinstance(param, ExtensionArray) or is_list_like(param):
                    ovalues = param
                else:  # Assume its an object
                    ovalues = [param] * len(self)
                return ovalues

            if isinstance(other, (ABCSeries, ABCIndexClass)):
                # rely on pandas to unbox and dispatch to us
                return NotImplemented

            if isinstance(other, LadybugArrayType):
                if self.dtype == other.dtype:
                    lvalues = self
                    rvalues = other
                else:

                    lvalues = self.data
                    rvalues = other.data

                    # Special case for comparison operations
                    # Values cannot be compared if they are
                    # of a different LadybugDType
                    if op.__name__ in COMPARISON_OPS:
                        res = [False]* len(lvalues)
                        return res

            else:
                lvalues = self.data
                rvalues = convert_values(other)


            # If the operator is not defined for the underlying objects,
            # a TypeError should be raised
            res = [op(a, b) for (a, b) in zip(lvalues, rvalues)]

            def _maybe_convert(arr):
                if coerce_to_dtype:
                    # https://github.com/pandas-dev/pandas/issues/22850
                    # We catch all regular exceptions here, and fall back
                    # to an ndarray.
                    if isinstance(rvalues, ExtensionArray) and rvalues.dtype != self.dtype:
                        res = np.asarray(arr)
                    else:
                        res = try_cast_to_ea(self, arr)
                    if not isinstance(res, type(self)):
                        # exception raised in _from_sequence; ensure we have ndarray
                        res = np.asarray(arr)
                else:
                    res = np.asarray(arr)
                return res

            if op.__name__ in {"divmod", "rdivmod"}:
                a, b = zip(*res)
                return _maybe_convert(a), _maybe_convert(b)

            return _maybe_convert(res)

        op_name = ops._get_op_name(op, True)
        return set_function_name(_binop, op_name, cls)


class LadybugArrayType(ExtensionArray, LadybugExtensionScalarOpsMixin):

    # _dtype = LadybugDtype()
    _dtype = LadybugDType()

    def __init__(self, values, dtype=None, copy=False):

        # self._dtype = dtype
        if dtype is not None:
            if isinstance(dtype, str):
                # Will raise error if not right
                self._dtype.construct_from_string(dtype)
            else:
                self._dtype = dtype

        if isinstance(dtype, ExtensionDtype) or dtype is None:
            values = np.asarray(values, dtype=np.float_)
        elif isinstance(values, LadybugArrayType):
            values = values.data
            self._dtype = values.dtype
        else:
            values = np.asarray(values, dtype=dtype)

        # TODO: dtype?
        if copy:
            values = values.copy()

        self.data = values
        

    @classmethod
    def _from_sequence(cls, scalars: Sequence, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of scalars.
        Parameters
        ----------
        scalars : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.
        Returns
        -------
        ExtensionArray
        """
        if dtype is None:
            if isinstance(scalars, cls):
                dtype = scalars.dtype
            elif isinstance(scalars, list) and isinstance(scalars[0], np.bool_):
                dtype = np.dtype("bool")
            else:
                dtype = None

        return cls(scalars, dtype=dtype, copy=copy)


    @classmethod
    def _from_sequence_of_strings(cls, strings, dtype=None, copy=False):
        """
        Construct a new ExtensionArray from a sequence of strings.
        .. versionadded:: 0.24.0
        Parameters
        ----------
        strings : Sequence
            Each element will be an instance of the scalar type for this
            array, ``cls.dtype.type``.
        dtype : dtype, optional
            Construct for this particular dtype. This should be a Dtype
            compatible with the ExtensionArray.
        copy : bool, default False
            If True, copy the underlying data.
        Returns
        -------
        ExtensionArray
        """
        return cls([float(string) for string in strings], dtype=dtype, copy=copy)

    @classmethod
    def _from_factorized(cls, values, original):
        """
        Reconstruct an ExtensionArray after factorization.
        Parameters
        ----------
        values : ndarray
            An integer ndarray with the factorized values.
        original : ExtensionArray
            The original ExtensionArray that factorize was called on.
        See Also
        --------
        factorize
        ExtensionArray.factorize
        """
        dtype = None

        if isinstance(original, cls):
            dtype = original.dtype
        elif isinstance(original, list):
            dtype = original[0].dtype
        else:
            raise f'Original value of type {type(original)} not supported'

        return cls(values, dtype)
        

    @classmethod
    def _from_data_collection(cls, datacollection: BaseCollection):
        """Generate a Ladybug Array from a Ladybug DataCollection

        Arguments:
            datacollection {BaseCollection} -- A ladybug data collection

        Returns:
            LadybugArrayType -- A Ladybug Array
        """
        dtype =  LadybugDType.construct_from_header(datacollection.header)

        return cls(values=datacollection.values, dtype=dtype)

    
    # def to_hourly_discontinuous_collection(self, datetimes: List[DateTime]) -> HourlyDiscontinuousCollection:
    #     assert len(self.data) == len(datetimes), f'Datetime index of length {len(datetimes)} does not match length of array {len(self)}'

    #     return HourlyDiscontinuousCollection(
    #         header=self.dtype.to_header(),
    #         values=self.data.tolist(),
    #         datetimes=datetimes,    
    #     )


    def convert_to_unit(self, unit: str):
        """Convert the Data Collection to the input unit

        Arguments:
            unit {str} -- A unit string that exists for the data type
        """
        self.data = self.dtype.data_type.to_unit(
            self.data, unit, self.dtype.unit)
        self.dtype.unit = unit

    def convert_to_ip(self):
        """Convert the Data Collection to IP units."""
        self.data, self.dtype.unit = self.dtype.data_type.to_ip(
                self.data, self.dtype.unit)

    def convert_to_si(self):
        """Convert the Data Collection to SI units."""
        self.data, self.dtype.unit = self.dtype.data_type.to_si(
                self.data, self.dtype.unit)

    def to_unit(self, unit):
        """Return a Data Collection in the input unit
        
        Arguments:
            unit {str} -- A unit string that exists for the data type

        Returns:
            LadybugArrayType -- A Ladybug Array converted to the desired unit
        """
        new_data_c = self.copy()
        new_data_c.convert_to_unit(unit)
        return new_data_c

    def to_ip(self):
        """Return a Data Collection in IP units

        Returns:
            LadybugArrayType -- A Ladybug Array converted to the desired unit
        """
        new_data_c = self.copy()
        new_data_c.convert_to_ip()
        return new_data_c

    def to_si(self):
        """Return a Data Collection in SI units
        
        Returns:
            LadybugArrayType -- A Ladybug Array converted to the desired unit
        """
        new_data_c = self.copy()
        new_data_c.convert_to_si()
        return new_data_c


    def __getitem__(self, item):
        # type (Any) -> Any
        """
        Select a subset of self.
        Parameters
        ----------
        item : int, slice, or ndarray
            * int: The position in 'self' to get.
            * slice: A slice object, where 'start', 'stop', and 'step' are
              integers or None
            * ndarray: A 1-d boolean NumPy ndarray the same length as 'self'
        Returns
        -------
        item : scalar or ExtensionArray
        Notes
        -----
        For scalar ``item``, return a scalar value suitable for the array's
        type. This should be an instance of ``self.dtype.type``.
        For slice ``key``, return an instance of ``ExtensionArray``, even
        if the slice is length 0 or 1.
        For a boolean mask, return an instance of ``ExtensionArray``, filtered
        to the values where ``item`` is True.
        """
        if isinstance(item, int):
            datum = self.data[item]
            return self.data[item]
        elif isinstance(item, slice):
            pass
        elif isinstance(item, np.ndarray):
            pass
        elif isinstance(item, list):
            try:
                item = np.asarray(item, dtype=np.int_)
            except Exception as e:        
                raise ValueError("Cannot index with an integer indexer containing NA values")

        elif isinstance(item, pd.core.arrays.boolean.BooleanArray):
            item = item.to_numpy(dtype="bool", na_value=False) 
        elif isinstance(item, pd.core.arrays.integer.IntegerArray):
            try:
                item = item.to_numpy(dtype="int", na_value=pd.NA)
            except Exception as e:
                raise ValueError("Cannot index with an integer indexer containing NA values")
        else:
            raise IndexError(f'Item type note recognised {type(item)}')

        view = self.view()
        view.data = view.data[item]

        return view


    def __len__(self) -> int:
        """
        Length of this array
        Returns
        -------
        length : int
        """
        return self.data.size


    def __setitem__(self, key: Union[int, np.ndarray], value: Any) -> None:
        """
        Set one or more values inplace.
    
        This method is not required to satisfy the pandas extension array
        interface.
    
        Parameters
        ----------
        key : int, ndarray, or slice
            When called from, e.g. ``Series.__setitem__``, ``key`` will be
            one of
    
            * scalar int
            * ndarray of integers.
            * boolean ndarray
            * slice object
    
        value : ExtensionDtype.type, Sequence[ExtensionDtype.type], or object
            value or values to be set of ``key``.
    
        Returns
        -------
        None
        """
        # Some notes to the ExtensionArray implementor who may have ended up
        # here. While this method is not required for the interface, if you
        # *do* choose to implement __setitem__, then some semantics should be
        # observed:
        #
        # * Setting multiple values : ExtensionArrays should support setting
        #   multiple values at once, 'key' will be a sequence of integers and
        #  'value' will be a same-length sequence.
        #
        # * Broadcasting : For a sequence 'key' and a scalar 'value',
        #   each position in 'key' should be set to 'value'.
        #
        # * Coercion : Most users will expect basic coercion to work. For
        #   example, a string like '2018-01-01' is coerced to a datetime
        #   when setting on a datetime64ns array. In general, if the
        #   __init__ method coerces that value, then so should __setitem__
        # Note, also, that Series/DataFrame.where internally use __setitem__
        # on a copy of the data.


        if isinstance(key, pd.core.arrays.BooleanArray):
            key = key.fillna(False)
            key = key.to_numpy(dtype="bool")
        elif isinstance(key, pd.core.arrays.IntegerArray):
            try:
                key = key.to_numpy(dtype="int")
            except ValueError as error:
                if 'Specify an appropriate \'na_value\'' in str(error):
                    raise ValueError('Cannot index with an integer indexer containing NA values')
                raise error

        elif isinstance(key, list):
            list_type = None
            list_item = 0

            while list_item < len(key):
                if isinstance(key[list_item], int):
                    list_type = "int"
                    break
                elif isinstance(key[list_item], bool):
                    list_type = "bool"
                    break
                else:
                    list_item += 1

            assert list_type is not None, IndexError('arrays used as indices must be of integer (or boolean) type')

            try:
                key = np.asarray(key, dtype=list_type)
            except TypeError as error:
                if 'int() argument must be a string, a bytes-like object or a number, not \'NAType\'' in str(error):
                    raise ValueError('Cannot index with an integer indexer containing NA values')

                raise error

            if list_type == 'bool':
                key = key.fillna(False)

        try:
            self.data[key] = value
        except IndexError as error:
            if "boolean index did not match indexed array along dimension" in str(error):
                raise IndexError('wrong length: ' + str(error))
            else:
                raise error

    
    # def __add__(self, other):
        
    #     if isinstance(other, self.__class__):
    #         self.data += other.data

    #     elif isinstance(other, (float, int, np.array)):
    #         self.data += other

    #     # elif isi

    @property
    def dtype(self) -> ExtensionDtype:
        """
        An instance of 'ExtensionDtype'.
        """
        return self._dtype
        # if self._dtype is None:
        #     return self.data.dtype
        # else:
        #     return self._dtype


    @property
    def nbytes(self) -> int:
        """
        The number of bytes needed to store this object in memory.
        """
        # If this is expensive to compute, return an approximate lower bound
        # on the number of bytes needed.
        # return self._itemsize * self.__len__
        return self.data.nbytes


    @property
    def _ndarray(self):
        return self.data

    def _reduce(self, name, skipna=True, **kwargs):
        """
        Return a scalar result of performing the reduction operation.
        Parameters
        ----------
        name : str
            Name of the function, supported values are:
            { any, all, min, max, sum, mean, median, prod,
            std, var, sem, kurt, skew }.
        skipna : bool, default True
            If True, skip NaN values.
        **kwargs
            Additional keyword arguments passed to the reduction function.
            Currently, `ddof` is the only supported kwarg.
        Returns
        -------
        scalar
        Raises
        ------
        TypeError : subclass does not define reductions
        """
        
        root = np

        child_kwargs = {}

        if 'ddof' in kwargs:
            child_kwargs['ddof'] = kwargs['ddof']

        if name == 'kurt':
            child_kwargs['bias'] = False
            name = 'kurtosis'
            root = stats

        if name == 'skew':
            child_kwargs['bias'] = False
            root = stats

        if skipna is True:
            # return getattr(self.data[~np.isnan(self.data)], name)(**child_kwargs)
            return getattr(root, name)(self.data[~np.isnan(self.data)], **child_kwargs)
        else:
            # return getattr(self.data, name)(**child_kwargs)
            return getattr(root, name)(self.data, **child_kwargs)


    def isna(self) -> ArrayLike:
        """
        A 1-D array indicating if each value is missing.
        Returns
        -------
        na_values : Union[np.ndarray, ExtensionArray]
            In most cases, this should return a NumPy ndarray. For
            exceptional cases like ``SparseArray``, where returning
            an ndarray would be expensive, an ExtensionArray may be
            returned.
        Notes
        -----
        If returning an ExtensionArray, then
        * ``na_values._is_boolean`` should be True
        * `na_values` should implement :func:`ExtensionArray._reduce`
        * ``na_values.any`` and ``na_values.all`` should be implemented
        """
        return np.isnan(self.data)


    def value_counts(self, dropna=False):
        """
        Return a Series containing counts of unique values.

        Parameters
        ----------
        dropna : bool, default True
            Don't include counts of NaT values.

        Returns
        -------
        Series
        """
        from pandas import Series, Index

        if dropna:
            values = self[~self.isna()].data
        else:
            values = self.data

        unique, counts = np.unique(values, return_counts=True)

        return Series(counts, index=unique, dtype="int64")


    def take(
        self, indices: Sequence[int], allow_fill: bool = False, fill_value: Any = None
    ) -> "ExtensionArray":
        """
        Take elements from an array.
        Parameters
        ----------
        indices : sequence of int
            Indices to be taken.
        allow_fill : bool, default False
            How to handle negative values in `indices`.
            * False: negative values in `indices` indicate positional indices
              from the right (the default). This is similar to
              :func:`numpy.take`.
            * True: negative values in `indices` indicate
              missing values. These values are set to `fill_value`. Any other
              other negative values raise a ``ValueError``.
        fill_value : any, optional
            Fill value to use for NA-indices when `allow_fill` is True.
            This may be ``None``, in which case the default NA value for
            the type, ``self.dtype.na_value``, is used.
            For many ExtensionArrays, there will be two representations of
            `fill_value`: a user-facing "boxed" scalar, and a low-level
            physical NA value. `fill_value` should be the user-facing version,
            and the implementation should handle translating that to the
            physical version for processing the take if necessary.
        Returns
        -------
        ExtensionArray
        Raises
        ------
        IndexError
            When the indices are out of bounds for the array.
        ValueError
            When `indices` contains negative values other than ``-1``
            and `allow_fill` is True.
        See Also
        --------
        numpy.take
        api.extensions.take
        Notes
        -----
        ExtensionArray.take is called by ``Series.__getitem__``, ``.loc``,
        ``iloc``, when `indices` is a sequence of values. Additionally,
        it's called by :meth:`Series.reindex`, or any other method
        that causes realignment, with a `fill_value`.
        Examples
        --------
        Here's an example implementation, which relies on casting the
        extension array to object dtype. This uses the helper method
        :func:`pandas.api.extensions.take`.
        .. code-block:: python
           def take(self, indices, allow_fill=False, fill_value=None):
               from pandas.core.algorithms import take
               # If the ExtensionArray is backed by an ndarray, then
               # just pass that here instead of coercing to object.
               data = self.astype(object)
               if allow_fill and fill_value is None:
                   fill_value = self.dtype.na_value
               # fill value should always be translated from the scalar
               # type for the array, to the physical storage type for
               # the data, before passing to take.
               result = take(data, indices, fill_value=fill_value,
                             allow_fill=allow_fill)
               return self._from_sequence(result, dtype=self.dtype)
        """
        # Implementer note: The `fill_value` parameter should be a user-facing
        # value, an instance of self.dtype.type. When passed `fill_value=None`,
        # the default of `self.dtype.na_value` should be used.
        # This may differ from the physical storage type your ExtensionArray
        # uses. In this case, your implementation is responsible for casting
        # the user-facing type to the storage type, before using
        # pandas.api.extensions.take
        copy = self.copy()
        copy.data = pd.api.extensions.take(copy.data, indices, allow_fill=allow_fill, fill_value=fill_value)
        
        return copy


    def copy(self) -> "ExtensionArray":
        """
        Return a copy of the array.
        Returns
        -------
        ExtensionArray
        """
        return self.__class__._from_factorized(self.data.copy(), self)


    def view(self, dtype=None) -> Union[ABCExtensionArray, np.ndarray]:
        """
        Return a view on the array.

        Parameters
        ----------
        dtype : str, np.dtype, or ExtensionDtype, optional
            Default None.

        Returns
        -------
        ExtensionArray
            A view of the :class:`ExtensionArray`.
        """
        # NB:
        # - This must return a *new* object referencing the same data, not self.
        # - The only case that *must* be implemented is with dtype=None,
        #   giving a view with the same dtype as self.
        if dtype is not None:
            raise NotImplementedError(dtype)

        return self.__class__._from_factorized(self.data, self)



    @classmethod
    def _concat_same_type(
        cls, to_concat: Sequence["ExtensionArray"]
    ) -> "ExtensionArray":
        """
        Concatenate multiple array.
        Parameters
        ----------
        to_concat : sequence of this type
        Returns
        -------
        ExtensionArray
        """
        data = np.concatenate([ga.data for ga in to_concat])
        return cls._from_factorized(data, to_concat)


    def astype(self, dtype, copy=True):
        if isinstance(dtype, self.dtype.__class__):
            if copy:
                return self.copy()
            return self

        # return np.array(self.data, dtype=dtype, copy=copy)
        return super(LadybugArrayType, self).astype(dtype)


LadybugArrayType._add_arithmetic_ops()
LadybugArrayType._add_comparison_ops()


from typing import Union
import pandas as pd
from ladybug import psychrometrics
from ladybug.datatype import TYPESDICT

from ..extension_types.dtype import LadybugDType
from ..extension_types.arraytype import LadybugArrayType


@pd.api.extensions.register_dataframe_accessor("psychro")
class PsychrometricsAccessor:

    """A pandas Dataframe accessor to perform psychrometric calculation operations

        Use this "psychro" accessor and pass in the name of a column in your dataframe, 
        a Series or an int/float value to perform as psychormetric operation and return a Series.

    Warning:
        If using a Dataframe column or a Series you must ensure that it is an instance
        of a LadybugArrayType. This is because the Psychrometric accessor runs type checks
        to convert a data type to the right unit if necessary.

    Example:
        .. code-block:: python

        import ladybug_pandas as lbp
        from ladybug.epw import EPW

        epw_path = 'tests/assets/epw/tokyo.epw'

        epw = EPW(epw_path)

        df = lbp.dataframe_from_epw(epw)

        df['Humidity Ratio'] = df.psychro.humid_ratio_from_db_rh(
            db_temp='Dry Bulb Temperature',
            rel_humid='Relative Humidity'
        )

    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj


    def _build_psychro_function(self, function, input_kwargs, output_type, output_unit, **kwargs):

        input_df = pd.DataFrame()

        function_inputs = {}

        for k, v in input_kwargs.items():
            input_value = kwargs.get(k)
            if isinstance(input_value, str):
                series = self._obj[input_value]
                input_type = TYPESDICT[v['input_type']]
                assert isinstance(series.values.dtype.data_type, input_type), \
                    f'Cannot use array of type {series.values.dtype.data_type} for {k}. Must be a {v["input_type"]}'

                input_df[k] = series.values.to_unit(v['input_unit'])

            elif isinstance(input_value, float) or isinstance(input_value, int):
                function_inputs[k] = input_value


        values = []

        if len(input_df) == 0:
            values.append(function(**function_inputs)) 
        else:
            for _, row in input_df.iterrows():
                function_inputs.update(row.to_dict())
                values.append(function(**function_inputs))

        dtype = LadybugDType(TYPESDICT[output_type](), output_unit)

        return pd.Series(LadybugArrayType(values, dtype))


    def saturated_vapor_pressure(self, t_kelvin: Union[str, float, int, LadybugArrayType]) -> pd.Series:

        return self._build_psychro_function(
            function=psychrometrics.saturated_vapor_pressure,
            input_kwargs={
                't_kelvin': {
                    'input_type': 'Temperature',
                    'input_unit': 'K'
                },
            },
            output_type='Pressure',
            output_unit='Pa',
            t_kelvin=t_kelvin
        )


    def humid_ratio_from_db_rh(self, db_temp, rel_humid, b_press=101325):
        """Humidity ratio (kg water/kg air) from air temperature (C) and relative humidity (%).

        Args:
            db_temp: Dry bulb temperature (C).
            rel_humid: Relative humidity (%).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Humidity ratio (kg water/kg air).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 20

            [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
            functions to calculate thermodynamic properties of air. Journal of
            Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
            https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
        """
        return self._build_psychro_function(
            function=psychrometrics.humid_ratio_from_db_rh,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'rel_humid': {
                    'input_type': 'RelativeHumidity',
                    'input_unit': '%'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='HumidityRatio',
            output_unit='%',
            db_temp=db_temp,
            rel_humid=rel_humid,
            b_press=b_press,
        )


    def enthalpy_from_db_hr(self, db_temp, humid_ratio, reference_temp=0):
        """Enthalpy (kJ/kg) at a given humidity ratio (water/air) and dry bulb temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            humid_ratio: Humidity ratio (kg water/kg air).
            reference_temp: Reference dry air temperature (C). Default is 0C,
                which is standard practice for SI enthalpy values. However, for
                IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
                aboslute thermodynamic enthalpy, one can input 0K (-273.15).

        Returns:
            Enthalpy (kJ/kg).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 30

            [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
            functions to calculate thermodynamic properties of air. Journal of
            Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
            https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
        """
        return self._build_psychro_function(
            function=psychrometrics.enthalpy_from_db_hr,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'humid_ratio': {
                    'input_type': 'HumidityRatio',
                    'input_unit': '%'
                },
                'reference_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                }
            },
            output_type='Enthalpy',
            output_unit='kJ/kg',
            db_temp=db_temp,
            humid_ratio=humid_ratio,
            reference_temp=reference_temp,
        )

    def dew_point_from_db_rh(self, db_temp, rel_humid):
        """Dew point temperature (C) from air temperature (C) and relative humidity (%).

        The dew point temperature is solved by inverting the equation giving water vapor
        pressure at saturation from temperature, which is relatively slow but ensures
        high accuracy down to 0.1 C at a wide range of dry bulb temperatures.
        The Newton-Raphson (NR) method is used on the logarithm of water vapour
        pressure as a function of temperature, which is a very smooth function
        Convergence is usually achieved in 3 to 5 iterations.

        Args:
            db_temp: Dry bulb temperature (C).
            rel_humid: Relative humidity (%).

        Returns:
            Dew point temperature (C).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn. 5 and 6

            [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
            functions to calculate thermodynamic properties of air. Journal of
            Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
            https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_rh,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'rel_humid': {
                    'input_type': 'RelativeHumidity',
                    'input_unit': '%'
                }
            },
            output_type='DewPointTemperature',
            output_unit='C',
            db_temp=db_temp,
            rel_humid=rel_humid,
        )


    def wet_bulb_from_db_rh(self, db_temp, rel_humid, b_press=101325):
        """Wet bulb temperature (C) from air temperature (C) and relative humidity (%).

        Args:
            db_temp: Dry bulb temperature (C).
            rel_humid: Relative humidity (%).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Wet bulb temperature (C).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 33 and 35

            [2] Meyer et al., (2019). PsychroLib: a library of psychrometric
            functions to calculate thermodynamic properties of air. Journal of
            Open Source Software, 4(33), 1137, https://doi.org/10.21105/joss.01137
            https://github.com/psychrometrics/psychrolib/blob/master/src/python/psychrolib.py
        """
        return self._build_psychro_function(
            function=psychrometrics.wet_bulb_from_db_rh,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'rel_humid': {
                    'input_type': 'RelativeHumidity',
                    'input_unit': '%'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='WetBulbTemperature',
            output_unit='C',
            db_temp=db_temp,
            rel_humid=rel_humid,
            b_press=b_press,
        )

    def rel_humid_from_db_hr(self, db_temp, humid_ratio, b_press=101325):
        """Relative Humidity (%) from humidity ratio (water/air) and air temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            humid_ratio: Humidity ratio (kg water/kg air).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Relative humidity (%).
        """
        return self._build_psychro_function(
            function=psychrometrics.rel_humid_from_db_hr,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'humid_ratio': {
                    'input_type': 'HumidityRatio',
                    'input_unit': '%'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='RelativeHumidity',
            output_unit='%',
            db_temp=db_temp,
            humid_ratio=humid_ratio,
            b_press=b_press,
        )


    def rel_humid_from_db_enth(self, db_temp, enthalpy, b_press=101325, reference_temp=0):
        """Relative Humidity (%) from air temperature (C) and enthalpy (kJ/kg).

        Args:
            db_temp: Dry bulb temperature (C).
            enthalpy: Enthalpy (kJ/kg).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).
            reference_temp: Reference dry air temperature (C). Default is 0C,
                which is standard practice for SI enthalpy values. However, for
                IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
                aboslute thermodynamic enthalpy, one can input 0K (-273.15).

        Returns:
            Relative humidity (%).
        """
        return self._build_psychro_function(
            function=psychrometrics.rel_humid_from_db_enth,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'enthalpy': {
                    'input_type': 'Enthalpy',
                    'input_unit': 'kJ/kg'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                },
                'reference_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                }
            },
            output_type='RelativeHumidity',
            output_unit='%',
            db_temp=db_temp,
            enthalpy=enthalpy,
            b_press=b_press,
            reference_temp=reference_temp,
        )


    def rel_humid_from_db_dpt(self, db_temp, dew_pt):
        """Relative humidity (%) from dry bulb temperature (C), and dew point temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            dew_pt: Dew point temperature (C).

        Returns:
            Relative humidity (%).
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_rh,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'dew_pt': {
                    'input_type': 'DewPointTemperature',
                    'input_unit': 'C'
                }
            },
            output_type='RelativeHumidity',
            output_unit='%',
            db_temp=db_temp,
            dew_pt=dew_pt,
        )


    def rel_humid_from_db_wb(self, db_temp, wet_bulb, b_press=101325):
        """Relative humidity (%) from dry bulb temperature (C), and wet bulb temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            wet_bulb: Wet bulb temperature (C).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Relative humidity (%).
        """
        return self._build_psychro_function(
            function=psychrometrics.rel_humid_from_db_wb,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'wet_bulb': {
                    'input_type': 'WetBulbTemperature',
                    'input_unit': 'C'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='RelativeHumidity',
            output_unit='%',
            db_temp=db_temp,
            wet_bulb=wet_bulb,
            b_press=b_press,
        )


    def dew_point_from_db_hr(self, db_temp, humid_ratio, b_press=101325):
        """Dew Point Temperature (C) from air temperature (C) and humidity ratio (water/air).

        Args:
            db_temp: Dry bulb temperature (C).
            humid_ratio: Humidity ratio (kg water/kg air).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Dew point temperature (C).
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_hr,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'humid_ratio': {
                    'input_type': 'HumidityRatio',
                    'input_unit': '%'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='DewPointTemperature',
            output_unit='C',
            db_temp=db_temp,
            humid_ratio=humid_ratio,
            b_press=b_press,
        )

    def dew_point_from_db_enth(self, db_temp, enthalpy, b_press=101325, reference_temp=0):
        """Dew point temperature (C) from air temperature (C) and enthalpy (kJ/kg).

        Args:
            db_temp: Dry bulb temperature (C).
            enthalpy: Enthalpy (kJ/kg).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).
            reference_temp: Reference dry air temperature (C). Default is 0C,
                which is standard practice for SI enthalpy values. However, for
                IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
                aboslute thermodynamic enthalpy, one can input 0K (-273.15).

        Returns:
            Dew point temperature (C).
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_enth,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'enthalpy': {
                    'input_type': 'Enthalpy',
                    'input_unit': 'kJ/kg'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                },
                'reference_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                }
            },
            output_type='DewPointTemperature',
            output_unit='C',
            db_temp=db_temp,
            enthlpy=enthalpy,
            b_press=b_press,
            reference_temp=reference_temp,
        )


    def dew_point_from_db_wb(self, db_temp, wet_bulb, b_press=101325):
        """Dew point temperature (C) from dry bulb (C) and wet bulb temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            wet_bulb: Wet bulb temperature (C).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Dew point temperature (C).
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_wb,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'wet_bulb': {
                    'input_type': 'WetBulbTemperature',
                    'input_unit': 'C'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='DewPointTemperature',
            output_unit='C',
            db_temp=db_temp,
            wet_bulb=wet_bulb,
            b_press=b_press,
        )


    def humid_ratio_from_db_wb(self, db_temp, wb_temp, b_press=101325):
        """Humidity ratio from air temperature (C) and wet bulb temperature (C).

        Args:
            db_temp: Dry bulb temperature (C).
            wb_temp: Wet bulb temperature (C).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Humidity ratio (kg water / kg air).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 36, solved for W
        """
        return self._build_psychro_function(
            function=psychrometrics.humid_ratio_from_db_wb,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'wb_temp': {
                    'input_type': 'WetBulbTemperature',
                    'input_unit': 'C'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='HumidityRatio',
            output_unit='%',
            db_temp=db_temp,
            wb_temp=wb_temp,
            b_press=b_press,
        )


    def db_temp_from_enth_hr(self, enthalpy, humid_ratio, reference_temp=0):
        """Dry bulb temperature (C) from enthalpy (kJ/kg) and humidity ratio (water/air).

        Args:
            enthalpy: Enthalpy (kJ/kg).
            humid_ratio: Humidity ratio (kg water/kg air).
            reference_temp: Reference dry air temperature (C). Default is 0C,
                which is standard practice for SI enthalpy values. However, for
                IP enthalpy, this is typically at 0F (-17.78C). Alternatively, for
                aboslute thermodynamic enthalpy, one can input 0K (-273.15).

        Returns:
            Dry bulb temperature (C).

        Note:
            [1] ASHRAE Handbook - Fundamentals (2017) ch. 1 eqn 30
        """
        return self._build_psychro_function(
            function=psychrometrics.db_temp_from_enth_hr,
            input_kwargs={
                'enthalpy': {
                    'input_type': 'Enthalpy',
                    'input_unit': 'kJ/kg'
                },
                'humid_ratio': {
                    'input_type': 'HumidityRatio',
                    'input_unit': '%'
                },
                'reference_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                }
            },
            output_type='DryBulbTemperature',
            output_unit='C',
            enthalpy=enthalpy,
            humid_ratio=humid_ratio,
            reference_temp=reference_temp,
        )


    def dew_point_from_db_rh_fast(self, db_temp, rel_humid):
        """Dew point temperature (C) from air temperature (C) and relative humidity (%).

        Note that the formula here is fast but is only accurate up to 90C. For accurate
        values at extreme temperatures, the dew_point_from_db_rh
        function should be used.

        Args:
            db_temp: Dry bulb temperature (C).
            rel_humid: Relative humidity (%).

        Returns:
            Dew point temperature (C).

        Note:
            [1] J. Sullivan and L. D. Sanders. "Method for obtaining wet-bulb temperatures
            by modifying the psychrometric formula." Center for Experiment Design and Data
            Analysis. NOAA - National Oceanic and Atmospheric Administration.
            https://www.weather.gov/epz/wxcalc_rh
        """
        return self._build_psychro_function(
            function=psychrometrics.dew_point_from_db_rh_fast,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'rel_humid': {
                    'input_type': 'RelativeHumidity',
                    'input_unit': '%'
                },
            },
            output_type='DewPointTemperature',
            output_unit='C',
            db_temp=db_temp,
            rel_humid=rel_humid,
        )


    def wet_bulb_from_db_rh_fast(self, db_temp, rel_humid, b_press=101325):
        """Wet bulb temperature (C) from air temperature (C) and relative humidity (%).

        Note that the formula here is fast but is only accurate around temperatures
        of 20C and lower. For accurate values at extreme temperatures, the
        wet_bulb_from_db_rh function should be used.

        Args:
            db_temp: Dry bulb temperature (C).
            rel_humid: Relative humidity (%).
            b_press: Air pressure (Pa). Default is pressure at sea level (101325 Pa).

        Returns:
            Wet bulb temperature (C).

        Note:
            [1] J. Sullivan and L. D. Sanders. "Method for obtaining wet-bulb temperatures
            by modifying the psychrometric formula." Center for Experiment Design and Data
            Analysis. NOAA - National Oceanic and Atmospheric Administration.
            https://www.weather.gov/epz/wxcalc_rh
        """
        return self._build_psychro_function(
            function=psychrometrics.wet_bulb_from_db_rh_fast,
            input_kwargs={
                'db_temp': {
                    'input_type': 'DryBulbTemperature',
                    'input_unit': 'C'
                },
                'rel_humid': {
                    'input_type': 'RelativeHumidity',
                    'input_unit': '%'
                },
                'b_press': {
                    'input_type': 'Pressure',
                    'input_unit': 'Pa'
                }
            },
            output_type='WetBulbTemperature',
            output_unit='C',
            db_temp=db_temp,
            rel_humid=rel_humid,
            b_press=b_press,
        )

    def _d_ln_p_ws(self, db_temp):
        """Helper function returning the derivative of the natural log of the
        saturation vapor pressure as a function of dry-bulb temperature.

        Args:
            db_temp : Dry bulb temperature (C).
        Returns:
            Derivative of natural log of vapor pressure of saturated air in Pa.
        """
        return self._build_psychro_function(
            function=psychrometrics._d_ln_p_ws,
            input_kwargs={
                'db_temp': {
                    'input_type': 'WetBulbTemperature',
                    'input_unit': 'C'
                },
            },
            output_type='Pressure',
            output_unit='Pa',
            db_temp=db_temp,
            rel_humid=rel_humid,
            b_press=b_press,
        )
